from __future__ import annotations

import os
from pathlib import Path
from tempfile import gettempdir
from typing import Any, Dict

import click
from pydantic import BaseSettings, Extra, Field, validator

from .utils import Literal


def _get_default_config_dir() -> Path:
    return Path(click.get_app_dir('instawow'))


class BaseConfig(BaseSettings):
    def _build_values(self, init_kwargs: Dict[str, Any], _env_file: Any = None) -> Dict[str, Any]:
        # Prioritise env vars
        return {**init_kwargs, **self._build_environ(_env_file)}


class _Config(BaseConfig):
    config_dir: Path = Field(default_factory=_get_default_config_dir)
    addon_dir: Path
    temp_dir: Path = Path(gettempdir()) / 'instawow'
    game_flavour: Literal['retail', 'classic']
    auto_update_check: bool = True

    class Config:
        env_prefix = 'INSTAWOW_'
        extra = Extra.allow

    @validator('config_dir', 'addon_dir', 'temp_dir')
    def _expand_paths(cls, value: Path) -> Path:
        try:
            return value.expanduser().resolve()
        except RuntimeError as error:
            # pathlib will raise RuntimeError for non-existent ~users
            raise ValueError(str(error)) from error

    @validator('addon_dir')
    def _check_writable(cls, value: Path) -> Path:
        if not (value.is_dir() and os.access(value, os.W_OK)):
            raise ValueError('must be a writable directory')
        return value

    @classmethod
    def read(cls) -> _Config:
        dummy_config = cls(addon_dir='', game_flavour='retail')
        return cls.parse_raw(dummy_config.config_file.read_text(encoding='utf-8'))

    def ensure_dirs(self) -> _Config:
        self.config_dir.mkdir(exist_ok=True, parents=True)
        for dir_ in self.logger_dir, self.plugin_dir, self.temp_dir, self.cache_dir:
            dir_.mkdir(exist_ok=True)
        return self

    def write(self) -> _Config:
        # This is separate from ``ensure_dirs`` because configuration
        # values are considered to be transient if overridden by an
        # env var, e.g. to bypass the cache
        self.ensure_dirs()
        output = self.json(include={'addon_dir', 'game_flavour'}, indent=2)
        self.config_file.write_text(output, encoding='utf-8')
        return self

    @property
    def is_classic(self) -> bool:
        return self.game_flavour == 'classic'

    @property
    def is_retail(self) -> bool:
        return self.game_flavour == 'retail'

    @property
    def config_file(self) -> Path:
        return self.config_dir / 'config.json'

    @property
    def logger_dir(self) -> Path:
        return self.config_dir / 'logs'

    @property
    def plugin_dir(self) -> Path:
        return self.config_dir / 'plugins'

    @property
    def cache_dir(self) -> Path:
        return self.temp_dir / 'cache'


Config = _Config



from __future__ import annotations

from datetime import datetime
import enum
from itertools import count, takewhile
import re
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterable,
    ClassVar,
    Dict,
    List,
    NamedTuple,
    Optional as O,
    Set,
    cast,
)

from pydantic import BaseModel
from yarl import URL

from . import exceptions as E, models as m
from .utils import (
    Literal,
    ManagerAttrAccessMixin,
    cached_property,
    gather,
    run_in_thread as t,
    slugify,
    uniq,
)

if TYPE_CHECKING:
    from .manager import Manager

    JsonDict = Dict[str, Any]


class Strategies(enum.Enum):
    default = enum.auto()
    latest = enum.auto()
    curse_latest_beta = enum.auto()
    curse_latest_alpha = enum.auto()
    any_flavour = enum.auto()


class Defn(NamedTuple):
    source: str
    name: str
    strategy: Strategies = Strategies.default
    source_id: O[str] = None

    @classmethod
    def from_pkg(cls, pkg: m.Pkg) -> Defn:
        return cls(
            pkg.source,
            pkg.slug,
            Strategies[pkg.options.strategy],  # type: ignore
            pkg.id,
        )

    def with_name(self, name: str) -> Defn:
        return self.__class__(self.source, name, self.strategy, self.source_id)

    def with_strategy(self, strategy: Strategies) -> Defn:
        return self.__class__(self.source, self.name, strategy, self.source_id)

    def __str__(self) -> str:
        return f'{self.source}:{self.name}'


class _CItem(BaseModel):
    source: str
    id: str
    slug: str = ''
    name: str
    compatibility: List[Literal['retail', 'classic']]
    folders: List[List[str]] = []


class MasterCatalogue(BaseModel):
    __root__: List[_CItem]

    class Config:
        keep_untouched: Any = (cached_property,)

    @classmethod
    async def collate(cls) -> MasterCatalogue:
        from types import SimpleNamespace
        from .manager import Manager, init_web_client

        resolvers = (CurseResolver, WowiResolver, TukuiResolver)

        async with init_web_client() as web_client:
            faux_manager = cast(Manager, SimpleNamespace(web_client=web_client))
            items = [a for r in resolvers async for a in r(faux_manager).collect_items()]
        catalogue = cls(__root__=items)
        return catalogue

    @cached_property
    def curse_slugs(self) -> Dict[str, str]:
        return {a.slug: a.id for a in self.__root__ if a.source == 'curse'}


class Resolver(ManagerAttrAccessMixin):
    source: ClassVar[str]
    name: ClassVar[str]
    strategies: ClassVar[Set[Strategies]]

    def __init__(self, manager: Manager) -> None:
        self.manager = manager

    def __init_subclass__(cls) -> None:
        async def wrapper(self: Resolver, defn: Defn, *args: Any, **kwargs: Any) -> m.Pkg:
            if defn.strategy in self.strategies:
                return await resolve_one(self, defn, *args, **kwargs)
            raise E.PkgStrategyUnsupported(defn.strategy)

        resolve_one = cls.resolve_one
        cls.resolve_one = wrapper

    @staticmethod
    def decompose_url(value: str) -> O[str]:
        "Attempt to extract definition names from resolvable URLs."

    async def resolve(self, defns: List[Defn]) -> Dict[Defn, Any]:
        "Resolve add-on definitions into packages."
        raise NotImplementedError

    async def collect_items(self) -> AsyncIterable[_CItem]:
        "Yield add-ons from source for cataloguing."
        return
        yield


class CurseResolver(Resolver):
    source = 'curse'
    name = 'CurseForge'
    strategies = {
        Strategies.default,
        Strategies.latest,
        Strategies.curse_latest_beta,
        Strategies.curse_latest_alpha,
        Strategies.any_flavour,
    }

    # Reference: https://twitchappapi.docs.apiary.io/
    addon_api_url = URL('https://addons-ecs.forgesvc.net/api/v2/addon')

    @staticmethod
    def decompose_url(value: str) -> O[str]:
        url = URL(value)
        if url.host == 'www.wowace.com' and len(url.parts) > 2 and url.parts[1] == 'projects':
            return url.parts[2].lower()
        elif (
            url.host == 'www.curseforge.com'
            and len(url.parts) > 3
            and url.parts[1:3] == ('wow', 'addons')
        ):
            return url.parts[3].lower()

    async def resolve(self, defns: List[Defn]) -> Dict[Defn, Any]:
        ids_for_defns = {
            d: (d.source_id or self.catalogue.curse_slugs.get(d.name) or d.name) for d in defns
        }
        numeric_ids = {i for i in ids_for_defns.values() if i.isdigit()}
        async with self.web_client.post(self.addon_api_url, json=list(numeric_ids)) as response:
            if response.status == 404:
                json_response = []
            else:
                json_response = await response.json()

        api_results = {str(r['id']): r for r in json_response}
        results = await gather(
            self.resolve_one(d, api_results.get(i)) for d, i in ids_for_defns.items()
        )
        return dict(zip(defns, results))

    async def resolve_one(self, defn: Defn, metadata: O[JsonDict]) -> m.Pkg:
        if not metadata:
            raise E.PkgNonexistent

        files = metadata['latestFiles']
        if not files:
            raise E.PkgFileUnavailable('no files available for download')

        def is_not_libless(f: JsonDict):
            # There's also an 'isAlternate' field that's missing some
            # 50 lib-less files from c. 2008.  'exposeAsAlternative' is
            # absent from the /file endpoint
            return not f['exposeAsAlternative']

        if defn.strategy is Strategies.any_flavour:

            def is_compatible_with_game_version(f: JsonDict):
                return True

        else:
            classic_version_prefix = '1.13'
            flavour = 'wow_classic' if self.config.is_classic else 'wow_retail'

            def is_compatible_with_game_version(f: JsonDict):
                # Files can belong both to retail and classic
                # but ``gameVersionFlavor`` can only be one of
                # 'wow_retail' or 'wow_classic'.  To spice things up,
                # ``gameVersion`` might not be populated so we still have to
                # check the value of ``gameVersionFlavor``
                return f['gameVersionFlavor'] == flavour or any(
                    v.startswith(classic_version_prefix) is self.config.is_classic
                    for v in f['gameVersion']
                )

        # 1 = stable; 2 = beta; 3 = alpha
        if defn.strategy is Strategies.latest:

            def has_release_type(f: JsonDict):
                return True

        elif defn.strategy is Strategies.curse_latest_beta:

            def has_release_type(f: JsonDict):
                return f['releaseType'] == 2

        elif defn.strategy is Strategies.curse_latest_alpha:

            def has_release_type(f: JsonDict):
                return f['releaseType'] == 3

        else:

            def has_release_type(f: JsonDict):
                return f['releaseType'] == 1

        try:
            file = max(
                (
                    f
                    for f in files
                    if is_not_libless(f)
                    and is_compatible_with_game_version(f)
                    and has_release_type(f)
                ),
                # The ``id`` is just a counter so we don't have to
                # go digging around dates
                key=lambda f: f['id'],
            )
        except ValueError:
            raise E.PkgFileUnavailable(
                f'no files compatible with {self.config.game_flavour} '
                f'using {defn.strategy.name!r} strategy'
            )

        # 1 = embedded library
        # 2 = optional dependency
        # 3 = required dependency
        # 4 = tool
        # 5 = incompatible
        # 6 = include (wat)
        deps = [m.PkgDep(id=d['addonId']) for d in file['dependencies'] if d['type'] == 3]

        return m.Pkg(
            source=self.source,
            id=metadata['id'],
            slug=metadata['slug'],
            name=metadata['name'],
            description=metadata['summary'],
            url=metadata['websiteUrl'],
            file_id=file['id'],
            download_url=file['downloadUrl'],
            date_published=file['fileDate'],
            version=file['displayName'],
            options=m.PkgOptions(strategy=defn.strategy.name),
            deps=deps,
        )

    async def collect_items(self) -> AsyncIterable[_CItem]:
        classic_version_prefix = '1.13'
        flavours = ('retail', 'classic')

        def excise_compatibility(files: Any):
            for c in flavours:
                if any(f['gameVersionFlavor'] == f'wow_{c}' for f in files):
                    yield c
                elif any(
                    v.startswith(classic_version_prefix) is (c == 'classic')
                    for f in files
                    for v in f['gameVersion']
                ):
                    yield c

        step = 1000
        for index in count(0, step):
            url = (self.addon_api_url / 'search').with_query(
                # fmt: off
                gameId='1',
                sort='3', # Alphabetical
                pageSize=step, index=index,
                # fmt: on
            )
            async with self.web_client.get(url) as response:
                json_response = await response.json()

            if not json_response:
                break
            for item in json_response:
                folders = uniq(
                    tuple(m['foldername'] for m in f['modules'])
                    for f in item['latestFiles']
                    if not f['exposeAsAlternative']
                )
                yield _CItem(
                    source=self.source,
                    id=item['id'],
                    slug=item['slug'],
                    name=item['name'],
                    folders=folders,
                    compatibility=list(excise_compatibility(item['latestFiles'])),
                )


class WowiResolver(Resolver):
    source = 'wowi'
    name = 'WoWInterface'
    strategies = {Strategies.default}

    # Reference: https://api.mmoui.com/v3/globalconfig.json
    # There's also a v4 API corresponding to the as yet unreleased Minion v4,
    # which is fair to assume is unstable.  They changed the naming scheme t
    # camelCase and some fields which were strings were converted to numbers.
    # Neither API provides access to classic files for multi-file add-ons and
    # 'UICompatibility' can't be relied on to enforce compatibility
    # in instawow.  The API appears to inherit the version of the _latest_
    # file to have been uploaded, which for multi-file add-ons can be the
    # classic version.  Hoooowever the download link always points to the
    # 'retail' version, which for single-file add-ons belonging to the
    # classic category would be an add-on for classic.
    list_api_url = 'https://api.mmoui.com/v3/game/WOW/filelist.json'
    details_api_url = URL('https://api.mmoui.com/v3/game/WOW/filedetails/')

    _files: O[JsonDict] = None

    @staticmethod
    def decompose_url(value: str) -> O[str]:
        url = URL(value)
        if (
            url.host in {'wowinterface.com', 'www.wowinterface.com'}
            and len(url.parts) == 3
            and url.parts[1] == 'downloads'
        ):
            if url.name == 'landing.php':
                id_ = url.query.get('fileid')
                if id_:
                    return id_
            elif url.name == 'fileinfo.php':
                id_ = url.query.get('id')
                if id_:
                    return id_
            else:
                match = re.match(r'^(?:download|info)(?P<id>\d+)', url.name)
                if match:
                    return match.group('id')

    async def resolve(self, defns: List[Defn]) -> Dict[Defn, Any]:
        if self._files is None:
            from .manager import cache_json_response

            files = await cache_json_response(
                self.manager,
                self.list_api_url,
                3600,
                label=f'Synchronising {self.source} catalogue',
            )
            self._files = {i['UID']: i for i in files}

        ids_for_defns = {d: ''.join(takewhile(str.isdigit, d.name)) for d in defns}
        numeric_ids = {i for i in ids_for_defns.values() if i.isdigit()}

        url = self.details_api_url / f'{",".join(numeric_ids)}.json'
        async with self.web_client.get(url) as response:
            if response.status == 404:
                json_response = []
            else:
                json_response = await response.json()

        api_results = {r['UID']: {**self._files[r['UID']], **r} for r in json_response}
        results = await gather(
            self.resolve_one(d, api_results.get(i)) for d, i in ids_for_defns.items()
        )
        return dict(zip(defns, results))

    async def resolve_one(self, defn: Defn, metadata: O[JsonDict]) -> m.Pkg:
        if not metadata:
            raise E.PkgNonexistent

        # Files "pending" are downloadable but do not have a checksum which
        # we use as a unique file identifier
        # TODO: investigate using a different field for updates
        if metadata['UIPending'] == '1':
            raise E.PkgFileUnavailable('new file awaiting approval')

        return m.Pkg(
            source=self.source,
            id=metadata['UID'],
            slug=slugify(f'{metadata["UID"]} {metadata["UIName"]}'),
            name=metadata['UIName'],
            description=metadata['UIDescription'],
            url=metadata['UIFileInfoURL'],
            file_id=metadata['UIMD5'],
            download_url=metadata['UIDownload'],
            date_published=metadata['UIDate'],
            version=metadata['UIVersion'],
            options=m.PkgOptions(strategy=defn.strategy.name),
        )

    async def collect_items(self) -> AsyncIterable[_CItem]:
        async with self.web_client.get(self.list_api_url) as response:
            json_response = await response.json()
        for item in json_response:
            yield _CItem(
                source=self.source,
                id=item['UID'],
                name=item['UIName'],
                folders=[item['UIDir']],
                compatibility=['retail', 'classic'],
            )


class TukuiResolver(Resolver):
    source = 'tukui'
    name = 'Tukui'
    strategies = {Strategies.default}

    api_url = URL('https://www.tukui.org/api.php')

    retail_uis = {'-1': 'tukui', '-2': 'elvui', 'tukui': 'tukui', 'elvui': 'elvui'}

    @staticmethod
    def decompose_url(value: str) -> O[str]:
        url = URL(value)
        if url.host == 'www.tukui.org' and url.path in {
            '/addons.php',
            '/classic-addons.php',
            '/download.php',
        }:
            name = url.query.get('id') or url.query.get('ui')
            if name:
                return name

    async def resolve(self, defns: List[Defn]) -> Dict[Defn, Any]:
        results = await gather(self.resolve_one(d) for d in defns)
        return dict(zip(defns, results))

    async def resolve_one(self, defn: Defn) -> m.Pkg:
        name = ui_name = self.retail_uis.get(defn.name)
        if not name:
            name = ''.join(takewhile('-'.__ne__, defn.name))

        if self.config.is_classic:
            query = 'classic-addon'
        elif ui_name:
            query = 'ui'
        else:
            query = 'addon'

        url = self.api_url.with_query({query: name})
        async with self.web_client.get(url) as response:
            if not response.content_length:
                raise E.PkgNonexistent
            addon = await response.json(content_type=None)  # text/html

        return m.Pkg(
            source=self.source,
            id=addon['id'],
            slug=ui_name or slugify(f'{addon["id"]} {addon["name"]}'),
            name=addon['name'],
            description=addon['small_desc'],
            url=addon['web_url'],
            file_id=addon['version'],
            download_url=addon['url'],
            date_published=datetime.fromisoformat(addon['lastupdate']),
            version=addon['version'],
            options=m.PkgOptions(strategy=defn.strategy.name),
        )

    async def collect_items(self) -> AsyncIterable[_CItem]:
        for query, param in [
            ('ui', 'tukui'),
            ('ui', 'elvui'),
            ('addons', 'all'),
            ('classic-addons', 'all'),
        ]:
            url = self.api_url.with_query({query: param})
            async with self.web_client.get(url) as response:
                metadata = await response.json(content_type=None)  # text/html

            if query == 'ui':
                yield _CItem(
                    source=self.source,
                    id=metadata['id'],
                    name=metadata['name'],
                    compatibility=('retail',),
                )
            else:
                compatibility = ('retail' if query == 'addons' else 'classic',)
                for item in metadata:
                    yield _CItem(
                        source=self.source,
                        id=item['id'],
                        name=item['name'],
                        compatibility=compatibility,
                    )


class InstawowResolver(Resolver):
    source = 'instawow'
    name = 'instawow'
    strategies = {Strategies.default}

    _addons = {('0', 'weakauras-companion'), ('1', 'weakauras-companion-autoupdate')}

    async def resolve(self, defns: List[Defn]) -> Dict[Defn, Any]:
        results = await gather(self.resolve_one(d) for d in defns)
        return dict(zip(defns, results))

    async def resolve_one(self, defn: Defn) -> m.Pkg:
        try:
            id_, slug = next(p for p in self._addons if defn.name in p)
        except StopIteration:
            raise E.PkgNonexistent

        from .wa_updater import WaCompanionBuilder, WaConfigError

        builder = WaCompanionBuilder(self.manager)
        if id_ == '1':
            try:
                await builder.build()
            except WaConfigError:
                raise E.PkgFileUnavailable('account named not provided')

        checksum = await t(builder.checksum)()
        return m.Pkg(
            source=self.source,
            id=id_,
            slug=slug,
            name='WeakAuras Companion',
            description='A WeakAuras Companion clone.',
            url='https://github.com/layday/instawow',
            file_id=checksum,
            download_url=builder.addon_file.as_uri(),
            date_published=datetime.now(),
            version=checksum[:7],
            options=m.PkgOptions(strategy=defn.strategy.name),
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, NamedTuple, Optional as O, Sequence

from loguru import logger
from pydantic import BaseModel, Extra, ValidationError, validator
from yarl import URL

from .config import BaseConfig
from .utils import ManagerAttrAccessMixin, bucketise, dict_chain, gather, run_in_thread as t

if TYPE_CHECKING:
    from .manager import Manager


metadata_api = URL('https://data.wago.io/api/check/weakauras')
raw_api = URL('https://data.wago.io/api/raw/encoded')


class BuilderConfig(BaseConfig):
    account: str

    class Config:
        env_prefix = 'WAC_'


class WaConfigError(Exception):
    pass


class AuraEntry(BaseModel):
    id: str
    uid: str
    parent: O[str]
    url: URL
    version: int
    semver: O[str]
    ignore_wago_update: bool = False

    class Config:
        arbitrary_types_allowed = True
        fields = {'ignore_wago_update': 'ignoreWagoUpdate'}

    @validator('url', pre=True)
    def _convert_url(cls, value: str) -> URL:
        return URL(value)


class ApiMetadata__Changelog(BaseModel):
    format: O[str]
    text: str = ''

    class Config:
        extra = Extra.forbid


class ApiMetadata(BaseModel):
    id: str
    name: str
    slug: str
    url: str
    created: str
    modified: str
    game: str
    fork_of: O[str]
    username: str
    version: int
    version_string: str
    changelog: ApiMetadata__Changelog
    region_type: O[str]

    class Config:
        extra = Extra.forbid
        fields = {
            'id': '_id',
            'fork_of': 'forkOf',
            'version_string': 'versionString',
            'region_type': 'regionType',
        }


class _OutdatedAuras(NamedTuple):
    slug: str
    existing_auras: List[AuraEntry]
    metadata: ApiMetadata
    import_string: str


class WaCompanionBuilder(ManagerAttrAccessMixin):
    """A WeakAuras Companion port for shellfolk."""

    def __init__(self, manager: Manager, account: O[str] = None) -> None:
        self.manager = manager

        self.out_dir = self.config.plugin_dir / __name__
        self.addon_file = self.out_dir / 'WeakAurasCompanion.zip'
        self.account = account

    def ensure_dirs(self) -> WaCompanionBuilder:
        self.out_dir.mkdir(exist_ok=True)
        return self

    @staticmethod
    def extract_auras(source: str) -> Dict[str, List[AuraEntry]]:
        import re
        from slpp import SLPP

        class WaParser(SLPP):
            def decode(self, text: Any) -> Any:
                if not text or not isinstance(text, str):
                    return

                text = text[text.find('= ') + 1 :]
                text = re.sub(r' -- \[\d+\]$', '', text, flags=re.M)
                self.text = text
                self.at, self.ch, self.depth = 0, '', 0
                self.len = len(text)
                self.next_chr()
                return self.value()  # type: ignore

        wa_table = WaParser().decode(source)
        raw_auras = (a for a in wa_table['displays'].values() if a.get('url'))
        aura_groups = bucketise(map(AuraEntry.parse_obj, raw_auras), key=lambda a: a.url.parts[1])
        return aura_groups

    def extract_installed_auras(self) -> Dict[str, List[AuraEntry]]:
        import time

        try:
            builder_config = BuilderConfig(account=self.account)
        except ValidationError as error:
            raise WaConfigError from error

        saved_vars = (
            self.config.addon_dir.parents[1]
            / 'WTF/Account'
            / builder_config.account
            / 'SavedVariables/WeakAuras.lua'
        )
        start = time.perf_counter()
        aura_groups = self.extract_auras(saved_vars.read_text(encoding='utf-8'))
        logger.debug(f'auras extracted in {time.perf_counter() - start}s')
        return aura_groups

    async def get_wago_aura_metadata(self, aura_ids: Sequence[str]) -> List[ApiMetadata]:
        url = metadata_api.with_query(ids=','.join(aura_ids))
        async with self.web_client.get(url) as response:
            metadata = await response.json()

        results = dict_chain(
            aura_ids, None, ((i.slug, i) for i in map(ApiMetadata.parse_obj, metadata))
        )
        return list(results.values())

    async def get_wago_aura_import_string(self, aura_id: str) -> str:
        async with self.web_client.get(raw_api.with_query(id=aura_id)) as response:
            return await response.text()

    async def get_outdated_auras(
        self, aura_groups: Dict[str, List[AuraEntry]]
    ) -> List[_OutdatedAuras]:
        metadata = await self.get_wago_aura_metadata(list(aura_groups))
        outdated = [
            ((s, w), r)
            for (s, w), r in zip(aura_groups.items(), metadata)
            if r and r.version > next((a for a in w if not a.parent), w[0]).version
        ]
        import_strings = await gather(
            (self.get_wago_aura_import_string(r.id) for _, r in outdated), False
        )
        return [_OutdatedAuras(s, w, r, i) for ((s, w), r), i in zip(outdated, import_strings)]

    def make_addon(self, outdated_auras: Sequence[_OutdatedAuras]) -> None:
        from jinja2 import Environment, FunctionLoader
        from zipfile import ZipFile, ZipInfo

        def loader(filename: str) -> str:
            from importlib.resources import read_text
            from . import wa_templates

            return read_text(wa_templates, filename)

        jinja_env = Environment(
            trim_blocks=True, lstrip_blocks=True, loader=FunctionLoader(loader)
        )

        with ZipFile(self.ensure_dirs().addon_file, 'w') as file:

            def write_tpl(filename: str, ctx: Dict[str, Any]) -> None:
                # We're not using a plain string as the first argument to
                # ``writestr`` 'cause then the timestamp is generated dynamically
                # making the build unreproducible
                zip_info = ZipInfo(filename=f'WeakAurasCompanion/{filename}')
                file.writestr(zip_info, jinja_env.get_template(filename).render(ctx))

            write_tpl(
                'data.lua',
                {
                    'was': [
                        (
                            o.metadata.slug,
                            {
                                'name': o.metadata.name,
                                'author': o.metadata.username,
                                'encoded': o.import_string,
                                'wagoVersion': o.metadata.version,
                                'wagoSemver': o.metadata.version_string,
                                'versionNote': o.metadata.changelog.text,
                            },
                        )
                        for o in outdated_auras
                    ],
                    'uids': [
                        (a.uid, a.url.parts[1]) for o in outdated_auras for a in o.existing_auras
                    ],
                    'ids': [
                        (a.id, a.url.parts[1]) for o in outdated_auras for a in o.existing_auras
                    ],
                    'stash': [],
                },
            )  # Send to WAC not supported - always empty
            write_tpl('init.lua', {})
            write_tpl(
                'WeakAurasCompanion.toc',
                {'interface': '11303' if self.config.is_classic else '80300'},
            )

    async def build(self) -> None:
        aura_groups = {
            k: v
            for k, v in (await t(self.extract_installed_auras)()).items()
            if not any(i.ignore_wago_update for i in v)
        }
        outdated_auras = aura_groups and await self.get_outdated_auras(aura_groups)
        await t(self.make_addon)(outdated_auras)

    def checksum(self) -> str:
        from hashlib import sha256

        return sha256(self.addon_file.read_bytes()).hexdigest()



from functools import lru_cache, partial
from pathlib import Path
import re

import pytest

from instawow.config import Config
from instawow.manager import Manager, init_web_client, prepare_db_session
from instawow.utils import get_version


def pytest_addoption(parser):
    parser.addoption('--instawow-no-mock', action='store_true')


def should_mock(fn):
    import inspect
    import warnings

    def wrapper(request):
        if request.config.getoption('--instawow-no-mock'):
            warnings.warn('not mocking')
            return None

        args = (request.getfixturevalue(p) for p in inspect.signature(fn).parameters)
        return fn(*args)

    return wrapper


@lru_cache(maxsize=None)
def read_fixture(filename):
    return (Path(__file__).parent / 'fixtures' / filename).read_bytes()


@lru_cache(maxsize=None)
def make_zip(name):
    from io import BytesIO
    from zipfile import ZipFile

    buffer = BytesIO()
    with ZipFile(buffer, 'w') as file:
        file.writestr(f'{name}/{name}.toc', b'')
    return buffer.getvalue()


@pytest.fixture(scope='session')
def temp_dir(tmp_path_factory):
    yield tmp_path_factory.mktemp('temp')


@pytest.fixture(params=['retail', 'classic'])
def partial_config(tmp_path, request, temp_dir):
    addons = tmp_path / 'addons'
    addons.mkdir()
    return {'addon_dir': addons, 'temp_dir': temp_dir, 'game_flavour': request.param}


@pytest.fixture
def full_config(tmp_path, partial_config):
    return {**partial_config, 'config_dir': tmp_path / 'config'}


@pytest.fixture
async def web_client():
    async with init_web_client() as web_client:
        yield web_client


@pytest.fixture
def manager(full_config, web_client):
    config = Config(**full_config).write()
    db_session = prepare_db_session(config=config)

    manager = Manager(config, db_session)
    manager.web_client = web_client
    yield manager


@pytest.fixture
def JsonResponse(aresponses):
    return partial(aresponses.Response, headers={'Content-Type': 'application/json'})


@pytest.fixture
def mock_pypi(aresponses, JsonResponse):
    aresponses.add(
        'pypi.org',
        '/pypi/instawow/json',
        'get',
        JsonResponse(body=f'{{"info": {{"version": "{get_version()}"}}}}'),
        repeat=float('inf'),
    )


@pytest.fixture
@should_mock
def mock_master_catalogue(aresponses, JsonResponse):
    aresponses.add(
        'raw.githubusercontent.com',
        aresponses.ANY,
        'get',
        JsonResponse(body=read_fixture('master-catalogue.json')),
        repeat=float('inf'),
    )


@pytest.fixture
@should_mock
def mock_curse(aresponses, JsonResponse, mock_master_catalogue):
    aresponses.add(
        'addons-ecs.forgesvc.net',
        '/api/v2/addon',
        'post',
        JsonResponse(body=read_fixture('curse-post-addon_all.json')),
        repeat=float('inf'),
    )
    aresponses.add(
        'edge.forgecdn.net',
        aresponses.ANY,
        'get',
        aresponses.Response(body=make_zip('Molinari')),
        repeat=float('inf'),
    )


@pytest.fixture
@should_mock
def mock_wowi(aresponses, JsonResponse, mock_master_catalogue):
    aresponses.add(
        'api.mmoui.com',
        '/v3/game/WOW/filelist.json',
        'get',
        JsonResponse(body=read_fixture('wowi-get-filelist.json')),
        repeat=float('inf'),
    )
    aresponses.add(
        'api.mmoui.com',
        re.compile(r'^/v3/game/WOW/filedetails/'),
        'get',
        JsonResponse(body=read_fixture('wowi-get-filedetails.json')),
        repeat=float('inf'),
    )
    aresponses.add(
        'cdn.wowinterface.com',
        aresponses.ANY,
        'get',
        aresponses.Response(body=make_zip('Molinari')),
        repeat=float('inf'),
    )


@pytest.fixture
@should_mock
def mock_tukui(aresponses, JsonResponse, mock_master_catalogue):
    aresponses.add(
        'www.tukui.org',
        '/api.php?ui=tukui',
        'get',
        JsonResponse(body=read_fixture('tukui-get-ui_tukui.json')),
        match_querystring=True,
        repeat=float('inf'),
    )
    aresponses.add(
        'www.tukui.org',
        '/api.php?addon=1',
        'get',
        JsonResponse(body=read_fixture('tukui-get-addon.json')),
        match_querystring=True,
        repeat=float('inf'),
    )
    aresponses.add(
        'www.tukui.org',
        '/api.php?classic-addon=1',
        'get',
        JsonResponse(body=read_fixture('tukui-get-classic-addon.json')),
        match_querystring=True,
        repeat=float('inf'),
    )
    aresponses.add('www.tukui.org', '/api.php', 'get', '', repeat=float('inf'))
    aresponses.add(
        'www.tukui.org',
        re.compile(r'^/downloads/tukui'),
        'get',
        aresponses.Response(body=make_zip('Tukui')),
        repeat=float('inf'),
    )
    aresponses.add(
        'www.tukui.org',
        '/addons.php?download=1',
        'get',
        aresponses.Response(body=make_zip('ElvUI_MerathilisUI')),
        match_querystring=True,
        repeat=float('inf'),
    )
    aresponses.add(
        'www.tukui.org',
        '/classic-addons.php?download=1',
        'get',
        aresponses.Response(body=make_zip('Tukui')),
        match_querystring=True,
        repeat=float('inf'),
    )


@pytest.fixture(autouse=True)
def mock_all(mock_pypi, mock_curse, mock_wowi, mock_tukui):
    pass



from __future__ import annotations

import asyncio
from collections import defaultdict, deque
from contextlib import contextmanager
from datetime import datetime, timedelta
from functools import partial
from itertools import chain, repeat
from pathlib import Path
import re
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Awaitable,
    Callable,
    Dict,
    Generic,
    Hashable,
    Iterable,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

try:
    from typing import Literal as _Literal
except ImportError:
    from typing_extensions import Literal as _Literal
Literal = _Literal  # ...

if TYPE_CHECKING:
    from prompt_toolkit.shortcuts import ProgressBar
    from .manager import CliManager

    _H = TypeVar('_H', bound=Hashable)
    _AnySet = TypeVar('_AnySet', bound=AbstractSet)
_V = TypeVar('_V')


class ManagerAttrAccessMixin:
    def __getattr__(self, name: str) -> Any:
        return getattr(self.manager, name)


class _TocEntry(NamedTuple):
    key: str
    value: str

    def __bool__(self) -> bool:
        return bool(self.value)


class TocReader:
    """Extracts key–value pairs from TOC files."""

    def __init__(self, contents: str) -> None:
        possible_entries = (
            map(str.strip, e.lstrip('#').partition(':')[::2])
            for e in contents.splitlines()
            if e.startswith('##')
        )
        self.entries = {k: v for k, v in possible_entries if k}
        self.default = ''

    def __getitem__(self, key: Union[str, Tuple[str, ...]]) -> _TocEntry:
        if isinstance(key, tuple):
            try:
                return next(filter(None, (self[k] for k in key)))
            except StopIteration:
                key = key[0]
        return _TocEntry(key, self.entries.get(key, self.default))

    @classmethod
    def from_path(cls, path: Path) -> TocReader:
        return cls(path.read_text(encoding='utf-8-sig', errors='replace'))

    @classmethod
    def from_path_name(cls, path: Path) -> TocReader:
        return cls.from_path(path / f'{path.name}.toc')


class cached_property(Generic[_V]):
    def __init__(self, f: Callable[[Any], _V]) -> None:
        self.f = f

    @overload
    def __get__(self, o: None, t: Optional[type] = None) -> Callable:
        ...

    @overload
    def __get__(self, o: Any, t: Optional[type] = None) -> _V:
        ...

    def __get__(self, o: Any, t: Optional[type] = None) -> Union[Callable, _V]:
        if o is None:
            return self.f
        else:
            o.__dict__[self.f.__name__] = v = self.f(o)
            return v


def bucketise(iterable: Iterable[_V], key: Callable[[_V], _H] = lambda v: v) -> Dict[_H, List[_V]]:
    "Place the elements of an iterable in a bucket according to ``key``."
    bucket: Any = defaultdict(list)
    for value in iterable:
        bucket[key(value)].append(value)
    return dict(bucket)


def dict_chain(
    keys: Iterable[_H], default: Any, *overrides: Iterable[Tuple[_H, Any]]
) -> Dict[_H, Any]:
    "Construct a dictionary from a series of iterables with overlapping keys."
    return dict(chain(zip(keys, repeat(default)), *overrides))


def uniq(it: Iterable[_H]) -> List[_H]:
    "Deduplicate hashable items in an iterable maintaining insertion order."
    return list(dict.fromkeys(it))


def merge_intersecting_sets(it: Iterable[_AnySet]) -> Iterable[_AnySet]:
    "Recursively merge intersecting sets in a collection."
    many_sets = deque(it)
    while True:
        try:
            this_set = many_sets.popleft()
        except IndexError:
            return
        while True:
            for other_set in many_sets:
                if this_set & other_set:
                    # The in-place operator will mutate unfrozen sets
                    # in the original collection
                    this_set = this_set | other_set  # type: ignore
                    many_sets.remove(other_set)
                    break
            else:
                break
        yield this_set


async def gather(it: Iterable[Awaitable[_V]], return_exceptions: bool = True) -> List[_V]:
    return await asyncio.gather(*it, return_exceptions=return_exceptions)


def run_in_thread(fn: Callable[..., _V]) -> Callable[..., Awaitable[_V]]:
    return lambda *a, **k: asyncio.get_running_loop().run_in_executor(None, lambda: fn(*a, **k))


@contextmanager
def copy_resources(*packages: str) -> Iterator[Path]:
    """Copy package resources to a temporary directory on disk.

    Alembic cannot construct a migration environment from memory.
    This is a genericised bodge to make migrations work in a frozen instawow.
    """
    from importlib.resources import contents, is_resource, read_binary
    from tempfile import TemporaryDirectory

    with TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)

        for package in packages:
            for resource in filter(partial(is_resource, package), contents(package)):
                parent_dir = tmp_path.joinpath(*package.split('.'))
                if not parent_dir.is_dir():
                    parent_dir.mkdir(parents=True)

                filename = parent_dir / resource
                # PyOxidizer does not expose Python source files to `importlib`
                # (see https://github.com/indygreg/PyOxidizer/issues/237).
                # Migrations have a "null" extension added to them that is shorn
                # from the extracted filename
                if filename.suffix == '.0':
                    filename = filename.with_suffix('')
                filename.write_bytes(read_binary(package, resource))
        yield tmp_path


def slugify(text: str) -> str:
    "Convert an add-on name into a lower-alphanumeric slug."
    return '-'.join(re.sub(r'[^0-9a-z]', ' ', text.casefold()).split())


def tabulate(rows: Sequence[Sequence[Any]], *, max_col_width: int = 60) -> str:
    "Produce an ASCII table from equal-length elements in a sequence."
    from textwrap import fill

    def apply_max_col_width(value: Sequence[Any]):
        return fill(str(value), width=max_col_width, max_lines=1)

    def calc_resultant_col_widths(rows: Sequence[Sequence[str]]):
        cols = zip(*rows)
        return [max(map(len, c)) for c in cols]

    rows = [tuple(map(apply_max_col_width, r)) for r in rows]
    head, *tail = rows

    base_template = ' '.join(f'{{{{{{0}}{w}}}}}' for w in calc_resultant_col_widths(rows))
    row_template = base_template.format(':<')
    table = '\n'.join(
        (
            base_template.format(':^').format(*head),
            base_template.format(f'0:-<').format(''),
            *(row_template.format(*r) for r in tail),
        )
    )
    return table


def make_progress_bar(**kwargs: Any) -> ProgressBar:
    "A ``ProgressBar`` with download progress expressed in megabytes."
    import signal
    from prompt_toolkit.formatted_text import HTML
    from prompt_toolkit.shortcuts.progress_bar import ProgressBar, formatters
    from prompt_toolkit.utils import Event

    # There is a race condition in the bar's shutdown logic
    # where the bar-drawing thread is left to run indefinitely
    # if the prompt_toolkit app is not (yet) running
    # but is scheduled to run when ``__exit__``ing the bar.
    # If an exception occurs early in the execution cycle or if execution
    # finishes before prompt_toolkit is able to crank the app,
    # instawow will hang.  This is my ham-fisted attempt to work
    # around all that by signalling to the daemon thread to kill the app.

    class PatchedProgressBar(ProgressBar):
        def __exit__(self, *args: Any):
            if self._has_sigwinch:
                self._loop.add_signal_handler(
                    signal.SIGWINCH, cast(Callable, self._previous_winch_handler)
                )

            if self._thread is not None:

                def attempt_exit(sender: Any):
                    sender.is_running and sender.exit()

                # Signal to ``_auto_refresh_context`` that it should exit the app
                self.app.on_invalidate = Event(self.app, attempt_exit)
                self._thread.join()

    class DownloadProgress(formatters.Progress):
        template = formatters.Progress.template + 'MB'

        def format(self, progress_bar: ProgressBar, progress: Any, width: int):
            def format_pct(value: int) -> str:
                return f'{value / 2 ** 20:.1f}'

            return HTML(self.template).format(
                current=format_pct(progress.current), total=format_pct(progress.total)
            )

    f = [
        formatters.Label(),
        formatters.Text(' '),
        formatters.Percentage(),
        formatters.Text(' '),
        formatters.Bar(),
        formatters.Text(' '),
        DownloadProgress(),
        formatters.Text(' '),
        formatters.Text('eta [', style='class:time-left'),
        formatters.TimeLeft(),
        formatters.Text(']', style='class:time-left'),
        formatters.Text(' '),
    ]
    progress_bar = PatchedProgressBar(formatters=f, **kwargs)
    return progress_bar


def shasum(*values: str) -> str:
    "Base-16-encode a string using SHA-256 truncated to 32 characters."
    from hashlib import sha256

    return sha256(''.join(values).encode()).hexdigest()[:32]


def is_not_stale(path: Path, ttl: int, unit: str = 'seconds') -> bool:
    "Check if a file is older than ``ttl``."
    mtime = path.exists() and path.stat().st_mtime
    return mtime > 0 and (
        (datetime.now() - datetime.fromtimestamp(cast(float, mtime))) < timedelta(**{unit: ttl})
    )


def get_version() -> str:
    "Get the installed version of instawow."
    try:
        from ._version import __version__
    except ImportError:
        return 'dev'
    else:
        return __version__


def is_outdated(manager: CliManager) -> bool:
    """Check on PyPI to see if the installed instawow is outdated.

    The response is cached for 24 hours.
    """
    __version__ = get_version()
    if 'dev' in __version__:
        return False

    def parse_version(version: str) -> Tuple[int, ...]:
        return tuple(map(int, version.split('.')[:3]))

    cache_file = manager.config.temp_dir / '.pypi_version'
    if is_not_stale(cache_file, 1, 'days'):
        version = cache_file.read_text(encoding='utf-8')
    else:
        from aiohttp.client import ClientError

        async def get_metadata():
            api_url = 'https://pypi.org/pypi/instawow/json'
            async with manager.web_client.get(api_url) as response:
                return await response.json()

        try:
            version = manager.run(get_metadata())['info']['version']
        except ClientError:
            version = __version__
        else:
            cache_file.write_text(version, encoding='utf-8')

    return parse_version(version) > parse_version(__version__)


def setup_logging(logger_dir: Path, level: Union[int, str] = 'INFO') -> int:
    from loguru import logger

    handler = {
        'sink': logger_dir / 'error.log',
        'level': level,
        'rotation': '1 MB',
        'enqueue': True,
    }
    (handler_id,) = logger.configure(handlers=(handler,))
    return handler_id



import json
from pathlib import Path
import sys

import pytest

from instawow.config import Config


def test_env_vars_have_prio(full_config, monkeypatch):
    monkeypatch.setenv('INSTAWOW_CONFIG_DIR', '/foo')
    monkeypatch.setenv('INSTAWOW_GAME_FLAVOUR', 'classic')

    config = Config(**full_config)
    assert config.config_dir == Path('/foo').resolve()
    assert config.game_flavour == 'classic'


def test_config_dir_is_populated(full_config):
    config = Config(**full_config).write()
    assert {i.name for i in config.config_dir.iterdir()} == {'config.json', 'logs', 'plugins'}


def test_reading_missing_config_from_env_raises(full_config, monkeypatch):
    monkeypatch.setenv('INSTAWOW_CONFIG_DIR', str(full_config['config_dir']))
    with pytest.raises(FileNotFoundError):  # type: ignore
        Config.read()


@pytest.mark.skipif(sys.platform == 'win32', reason='no ~ expansion on Windows')
@pytest.mark.parametrize('dir_', ['config_dir', 'addon_dir', 'temp_dir'])
def test_invalid_any_dir_raises(full_config, dir_):
    with pytest.raises(ValueError):  # type: ignore
        Config(**{**full_config, dir_: '~foo'})


def test_invalid_addon_dir_raises(full_config):
    with pytest.raises(ValueError, match='must be a writable directory'):  # type: ignore
        Config(**{**full_config, 'addon_dir': 'foo'})


def test_reading_config_file(full_config):
    Config(**full_config).write()
    config_json = {
        'addon_dir': str(full_config['addon_dir']),
        'game_flavour': full_config['game_flavour'],
    }
    assert config_json == json.loads((full_config['config_dir'] / 'config.json').read_text())


@pytest.mark.skipif(sys.platform == 'win32', reason='path handling')
def test_default_config_dir_is_platform_and_xdg_appropriate(partial_config, monkeypatch):
    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'linux')
        config_dir = Config(**partial_config).config_dir
        assert config_dir == Path.home() / '.config/instawow'

        patcher.setenv('XDG_CONFIG_HOME', '/foo')
        config_dir = Config(**partial_config).config_dir
        assert config_dir == Path('/foo/instawow')

    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'darwin')
        config_dir = Config(**partial_config).config_dir
        assert config_dir == Path.home() / 'Library/Application Support/instawow'


@pytest.mark.skipif(sys.platform != 'win32', reason='path unhandling')
def test_default_config_dir_on_win32(partial_config, monkeypatch):
    assert Config(**partial_config).config_dir == Path.home() / 'AppData/Roaming/instawow'
    monkeypatch.delenv('APPDATA')
    assert Config(**partial_config).config_dir == Path.home() / 'instawow'



from __future__ import annotations

from functools import partial
from itertools import chain, islice
from pathlib import Path
from textwrap import dedent, fill
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    FrozenSet,
    Generator,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

import click

from . import exceptions as E, models
from .resolvers import Defn, Strategies
from .utils import (
    TocReader,
    cached_property,
    get_version,
    is_outdated,
    setup_logging,
    tabulate,
    uniq,
)

if TYPE_CHECKING:
    from .manager import CliManager

    _R = TypeVar('_R')


class Report:
    _success = click.style('✓', fg='green')
    _failure = click.style('✗', fg='red')
    _warning = click.style('!', fg='blue')

    def __init__(
        self,
        results: Mapping[Defn, E.ManagerResult],
        filter_fn: Callable[[E.ManagerResult], bool] = lambda _: True,
    ):
        self.results = results
        self.filter_fn = filter_fn

    @property
    def code(self) -> int:
        return any(
            r
            for r in self.results.values()
            if isinstance(r, (E.ManagerError, E.InternalError)) and self.filter_fn(r)
        )

    def __str__(self) -> str:
        def _adorn_result(result: E.ManagerResult) -> str:
            if isinstance(result, E.InternalError):
                return self._warning
            elif isinstance(result, E.ManagerError):
                return self._failure
            return self._success

        return '\n'.join(
            f'{_adorn_result(r)} {click.style(str(a), bold=True)}\n'
            + fill(r.message, initial_indent='  ', subsequent_indent='  ')
            for a, r in self.results.items()
            if self.filter_fn(r)
        )

    def generate(self):
        manager: CliManager = click.get_current_context().obj.m
        if manager.config.auto_update_check and is_outdated(manager):
            click.echo(f'{self._warning} instawow is out of date')

        report = str(self)
        if report:
            click.echo(report)

    def generate_and_exit(self):
        self.generate()
        ctx = click.get_current_context()
        ctx.exit(self.code)


class ManagerWrapper:
    def __init__(self, debug: bool = False):
        self.debug = debug

    @cached_property
    def m(self) -> CliManager:
        import asyncio
        from .config import Config
        from .manager import CliManager, prepare_db_session

        # TODO: rm once https://github.com/aio-libs/aiohttp/issues/4324 is fixed
        policy = getattr(asyncio, 'WindowsSelectorEventLoopPolicy', None)
        if policy:
            asyncio.set_event_loop_policy(policy())

        while True:
            try:
                config = Config.read().ensure_dirs()
            except FileNotFoundError:
                ctx = click.get_current_context()
                ctx.invoke(configure)
            else:
                break

        setup_logging(config.logger_dir, 'DEBUG' if self.debug else 'INFO')
        db_session = prepare_db_session(config)
        manager = CliManager(config, db_session)
        return manager


M = ManagerWrapper


@overload
def parse_into_defn(manager: CliManager, value: str, *, raise_invalid: bool = True) -> Defn:
    ...


@overload
def parse_into_defn(
    manager: CliManager, value: Sequence[str], *, raise_invalid: bool = True
) -> List[Defn]:
    ...


def parse_into_defn(
    manager: CliManager, value: Sequence[str], *, raise_invalid: bool = True
) -> Union[Defn, List[Defn]]:
    if not isinstance(value, str):
        defns = (parse_into_defn(manager, v, raise_invalid=raise_invalid) for v in value)
        return uniq(defns)

    delim = ':'
    any_source = '*'
    if delim not in value:
        if raise_invalid:
            raise click.BadParameter(value)
        parts = (any_source, value)
    else:
        parts = manager.decompose_url(value)
        if not parts:
            parts = value.partition(delim)[::2]
    return Defn(*parts)


def parse_into_defn_with_strategy(
    manager: CliManager, value: Sequence[Tuple[str, str]]
) -> List[Defn]:
    defns = parse_into_defn(manager, [d for _, d in value])
    strategies = (Strategies[s] for s, _ in value)
    return list(map(Defn.with_strategy, defns, strategies))


def export_to_csv(pkgs: Sequence[models.Pkg], path: Path):
    "Export packages to CSV."
    import csv

    with Path(path).open('w', encoding='utf-8', newline='') as file:
        writer = csv.writer(file)
        writer.writerow(('defn', 'strategy'))
        writer.writerows((Defn.from_pkg(p), p.options.strategy) for p in pkgs)


def import_from_csv(manager: CliManager, path: Path) -> List[Defn]:
    "Import definitions from CSV."
    import csv

    with Path(path).open(encoding='utf-8', newline='') as contents:
        rows = [(b, a) for a, b in islice(csv.reader(contents), 1, None)]
    return parse_into_defn_with_strategy(manager, rows)


def _callbackify(fn: Callable[..., _R]) -> Callable[[click.Context, Any, Any], _R]:
    return lambda c, _, v: fn(c.obj.m, v)


def _combine_into(param_name: str, fn: Callable[..., Any]):
    def combine(ctx: click.Context, param: Any, value: Any):
        addons = ctx.params.setdefault(param_name, [])
        if value:
            addons.extend(fn(ctx.obj.m, value))

    return combine


def _show_version(ctx: click.Context, param: Any, value: bool):
    if value:
        __version__ = get_version()
        click.echo(f'instawow, version {__version__}')
        ctx.exit()


@click.group(context_settings={'help_option_names': ('-h', '--help')})
@click.option(
    '--version',
    is_flag=True,
    default=False,
    callback=_show_version,
    expose_value=False,
    is_eager=True,
    help='Show the version and exit.',
)
@click.option('--debug', is_flag=True, default=False, help='Log more things.')
@click.pass_context
def main(ctx: click.Context, debug: bool):
    "Add-on manager for World of Warcraft."
    if not ctx.obj:
        ctx.obj = ManagerWrapper(debug)


@main.command()
@click.option('--replace', '-o', is_flag=True, default=False, help='Replace existing add-ons.')
@click.option(
    '--import',
    '-i',
    type=click.Path(dir_okay=False, exists=True),
    callback=_combine_into('addons', import_from_csv),
    expose_value=False,
    help='Install add-ons from CSV.',
)
@click.option(
    '--with-strategy',
    '-s',
    multiple=True,
    type=(click.Choice([s.name for s in Strategies]), str),
    callback=_combine_into('addons', parse_into_defn_with_strategy),
    expose_value=False,
    metavar='<STRATEGY ADDON>...',
    help='A strategy followed by an add-on definition.  '
    'Available strategies are: '
    f'{", ".join(repr(s.name) for s in islice(Strategies, 1, None))}.',
)
@click.argument(
    'addons', nargs=-1, callback=_combine_into('addons', parse_into_defn), expose_value=False
)
@click.pass_obj
def install(obj: M, addons: Sequence[Defn], replace: bool):
    "Install add-ons."
    if not addons:
        raise click.UsageError(
            'You must provide at least one of "ADDONS", ' '"--with-strategy" or "--import"'
        )

    results = obj.m.run(obj.m.install(addons, replace))
    Report(results).generate_and_exit()


@main.command()
@click.argument('addons', nargs=-1, callback=_callbackify(parse_into_defn))
@click.pass_obj
def update(obj: M, addons: Sequence[Defn]):
    "Update installed add-ons."

    def report_filter(result: E.ManagerResult):
        # Hide package from output if up to date and ``update`` was invoked without args
        return cast(bool, addons or not isinstance(result, E.PkgUpToDate))

    defns = addons
    if not defns:
        defns = [Defn.from_pkg(p) for p in obj.m.db_session.query(models.Pkg).all()]

    results = obj.m.run(obj.m.update(defns))
    Report(results, report_filter).generate_and_exit()


@main.command()
@click.argument('addons', nargs=-1, required=True, callback=_callbackify(parse_into_defn))
@click.pass_obj
def remove(obj: M, addons: Sequence[Defn]):
    "Remove add-ons."
    results = obj.m.run(obj.m.remove(addons))
    Report(results).generate_and_exit()


@main.command()
@click.option(
    '--auto', '-a', is_flag=True, default=False, help='Do not ask for user confirmation.'
)
@click.pass_context
def reconcile(ctx: click.Context, auto: bool):
    "Reconcile pre-installed add-ons."
    from .matchers import AddonFolder, match_toc_ids, match_toc_names, match_dir_names, get_folders
    from .models import is_pkg
    from .prompts import PkgChoice, confirm, select, skip

    preamble = dedent(
        '''\
        Use the arrow keys to navigate, <o> to open an add-on in your browser,
        enter to make a selection and <s> to skip to the next item.

        Versions that differ from the installed version or differ between
        choices are highlighted in purple.

        The reconciler will perform three passes in decreasing order of accuracy,
        looking to match source IDs and add-on names in TOC files, and folders.

        Selected add-ons will be reinstalled.

        You can also run `reconcile` in promptless mode by passing
        the `--auto` flag.  In this mode, add-ons will be reconciled
        without user input.
        '''
    )

    manager: CliManager = ctx.obj.m

    def prompt_one(addons: List[AddonFolder], pkgs: List[models.Pkg]) -> Union[Defn, Tuple[()]]:
        def create_choice(pkg: models.Pkg):
            defn = Defn.from_pkg(pkg)
            title = [
                ('', str(defn)),
                ('', '=='),
                ('class:highlight-sub' if highlight_version else '', pkg.version),
            ]
            return PkgChoice(title, defn, pkg=pkg)

        # Highlight version if there's multiple of them
        highlight_version = len({i.version for i in chain(addons, pkgs)}) > 1
        choices = list(chain(map(create_choice, pkgs), (skip,)))
        addon = addons[0]
        # Using 'unsafe_ask' to let ^C bubble up
        selection = select(f'{addon.name} [{addon.version or "?"}]', choices).unsafe_ask()
        return selection

    def prompt(groups: Iterable[Tuple[List[AddonFolder], List[Defn]]]) -> Iterable[Defn]:
        results = manager.run(manager.resolve(list({d for _, b in groups for d in b})))
        for addons, defns in groups:
            shortlist = list(filter(is_pkg, (results[d] for d in defns)))
            if shortlist:
                if auto:
                    pkg = shortlist[0]  # TODO: something more sophisticated
                    selection = Defn.from_pkg(pkg)
                else:
                    selection = prompt_one(addons, shortlist)
                selection and (yield selection)

    def match_all() -> Generator[List[Defn], FrozenSet[AddonFolder], None]:
        # Match in order of increasing heuristicitivenessitude
        for fn in (
            match_toc_ids,
            match_dir_names,
            match_toc_names,
        ):
            groups = manager.run(fn(manager, (yield [])))
            yield list(prompt(groups))

    leftovers = get_folders(manager)
    if not leftovers:
        click.echo('No add-ons left to reconcile.')
        return
    if not auto:
        click.echo(preamble)

    matcher = match_all()
    for _ in matcher:  # Skip over consumer yields
        selections = matcher.send(leftovers)
        if selections and (auto or confirm('Install selected add-ons?').unsafe_ask()):
            results = manager.run(manager.install(selections, replace=True))
            Report(results).generate()

        leftovers = get_folders(manager)

    if not auto and leftovers:
        click.echo()
        table_rows = [('unreconciled',), *((f.name,) for f in sorted(leftovers))]
        click.echo(tabulate(table_rows))


@main.command()
@click.option(
    '--limit',
    '-l',
    default=10,
    type=click.IntRange(1, 20, clamp=True),
    help='A number to limit results to.',
)
@click.argument(
    'search-terms', nargs=-1, required=True, callback=lambda _, __, v: ' '.join(v)
)  # type: ignore
@click.pass_context
def search(ctx: click.Context, limit: int, search_terms: str):
    "Search for add-ons to install."
    from .prompts import PkgChoice, checkbox, confirm

    manager: CliManager = ctx.obj.m

    pkgs = manager.run(manager.search(search_terms, limit))
    if pkgs:
        choices = [PkgChoice(f'{p.name}  ({d}=={p.version})', d, pkg=p) for d, p in pkgs.items()]
        selections = checkbox('Select add-ons to install', choices=choices).unsafe_ask()
        if selections and confirm('Install selected add-ons?').unsafe_ask():
            ctx.invoke(install, addons=selections)


@main.command('list')
@click.option(
    '--export',
    '-e',
    default=None,
    type=click.Path(dir_okay=False, writable=True),
    help='Export listed add-ons to CSV.',
)
@click.option(
    '--format',
    '-f',
    'output_format',
    type=click.Choice(('simple', 'detailed', 'json')),
    default='simple',
    show_default=True,
    help='Change the output format.',
)
@click.argument(
    'addons', nargs=-1, callback=_callbackify(partial(parse_into_defn, raise_invalid=False))
)
@click.pass_obj
def list_(obj: M, addons: Sequence[Defn], export: Optional[str], output_format: str):
    "List installed add-ons."
    from sqlalchemy import and_, or_

    def format_deps(pkg: models.Pkg):
        deps = (Defn(pkg.source, d.id) for d in pkg.deps)
        deps = (d.with_name(getattr(obj.m.get(d), 'slug', d.name)) for d in deps)
        return map(str, deps)

    def get_desc(pkg: models.Pkg):
        if pkg.source == 'wowi':
            toc_reader = TocReader.from_path_name(obj.m.config.addon_dir / pkg.folders[0].name)
            return toc_reader['Notes'].value
        else:
            return pkg.description

    pkgs = (
        obj.m.db_session.query(models.Pkg)
        .filter(
            or_(
                *(
                    models.Pkg.slug.contains(d.name)
                    if d.source == '*'
                    else and_(
                        models.Pkg.source == d.source,
                        or_(models.Pkg.id == d.name, models.Pkg.slug == d.name),
                    )
                    for d in addons
                )
            )
        )
        .order_by(models.Pkg.source, models.Pkg.name)
        .all()
    )
    if export:
        export_to_csv(pkgs, cast(Path, export))
    elif pkgs:
        if output_format == 'json':
            from .models import MultiPkgModel

            click.echo(MultiPkgModel.from_orm(pkgs).json(indent=2))
        elif output_format == 'detailed':
            formatter = click.HelpFormatter(max_width=99)
            for pkg in pkgs:
                with formatter.section(Defn.from_pkg(pkg)):
                    formatter.write_dl(
                        (
                            ('Name', pkg.name),
                            ('Description', get_desc(pkg)),
                            ('URL', pkg.url),
                            ('Version', pkg.version),
                            ('Date published', pkg.date_published.isoformat(' ', 'minutes')),
                            ('Folders', ', '.join(f.name for f in pkg.folders)),
                            ('Dependencies', ', '.join(format_deps(pkg))),
                            ('Options', f'{{"strategy": "{pkg.options.strategy}"}}'),
                        )
                    )

            click.echo(formatter.getvalue(), nl=False)
        else:
            click.echo('\n'.join(str(Defn.from_pkg(p)) for p in pkgs))


@main.command()
@click.option(
    '--exclude-own', '-e', is_flag=True, default=False, help='Exclude folders managed by instawow.'
)
@click.option(
    '--toc-entry',
    '-t',
    'toc_entries',
    multiple=True,
    help='An entry to extract from the TOC.  Repeatable.',
)
@click.pass_obj
def list_folders(obj: M, exclude_own: bool, toc_entries: Sequence[str]) -> None:
    "List add-on folders."
    from .matchers import get_folders

    folders = sorted(get_folders(obj.m, exclude_own=exclude_own))
    if folders:
        rows = [
            ('unreconciled' if exclude_own else 'folder', *(f'[{e}]' for e in toc_entries)),
            *((f.name, *(f.toc_reader[e].value for e in toc_entries)) for f in folders),
        ]
        click.echo(tabulate(rows))


@main.command(hidden=True)
@click.argument('addon', callback=_callbackify(partial(parse_into_defn, raise_invalid=False)))
@click.pass_context
def info(ctx: click.Context, addon: Defn):
    "Alias of `list -f detailed`."
    ctx.invoke(list_, addons=(addon,), output_format='detailed')


@main.command()
@click.argument('addon', callback=_callbackify(partial(parse_into_defn, raise_invalid=False)))
@click.pass_obj
def visit(obj: M, addon: Defn):
    "Open an add-on's homepage in your browser."
    pkg = obj.m.get_from_substr(addon)
    if pkg:
        import webbrowser

        webbrowser.open(pkg.url)
    else:
        Report({addon: E.PkgNotInstalled()}).generate_and_exit()


@main.command()
@click.argument('addon', callback=_callbackify(partial(parse_into_defn, raise_invalid=False)))
@click.pass_obj
def reveal(obj: M, addon: Defn):
    "Open an add-on folder in your file manager."
    pkg = obj.m.get_from_substr(addon)
    if pkg:
        import webbrowser

        webbrowser.open((obj.m.config.addon_dir / pkg.folders[0].name).as_uri())
    else:
        Report({addon: E.PkgNotInstalled()}).generate_and_exit()


@main.command()
@click.option(
    '--active',
    'show_active',
    is_flag=True,
    default=False,
    help='Show the active configuration and exit.',
)
@click.pass_obj
def configure(obj: M, show_active: bool):
    "Configure instawow."
    if show_active:
        click.echo(obj.m.config.json(indent=2))
        return

    from prompt_toolkit.completion import WordCompleter
    from prompt_toolkit.shortcuts import prompt

    from .config import Config
    from .prompts import DirectoryCompleter, PydanticValidator

    addon_dir = prompt(
        'Add-on directory: ',
        completer=DirectoryCompleter(),
        validator=PydanticValidator(Config, 'addon_dir'),
    )
    game_flavour = prompt(
        'Game flavour: ',
        default='classic' if '_classic_' in addon_dir else 'retail',
        completer=WordCompleter(['retail', 'classic']),
        validator=PydanticValidator(Config, 'game_flavour'),
    )
    config = Config(addon_dir=addon_dir, game_flavour=game_flavour).write()
    click.echo(f'Configuration written to: {config.config_file}')


@main.group('weakauras-companion')
def _weakauras_group():
    "Manage your WeakAuras."


@_weakauras_group.command('build')
@click.option(
    '--account',
    '-a',
    required=True,
    help='Your account name.  This is used to locate ' 'the WeakAuras data file.',
)
@click.pass_obj
def build_weakauras_companion(obj: M, account: str):
    "Build the WeakAuras Companion add-on."
    from .wa_updater import WaCompanionBuilder

    obj.m.run(WaCompanionBuilder(obj.m, account).build())


@_weakauras_group.command('list')
@click.option(
    '--account',
    '-a',
    required=True,
    help='Your account name.  This is used to locate ' 'the WeakAuras data file.',
)
@click.pass_obj
def list_installed_wago_auras(obj: M, account: str):
    "List WeakAuras installed from Wago."
    from .wa_updater import WaCompanionBuilder

    aura_groups = WaCompanionBuilder(obj.m, account).extract_installed_auras()
    installed_auras = sorted(
        (a.id, a.url, 'yes' if a.ignore_wago_update else 'no')
        for v in aura_groups.values()
        for a in v
        if not a.parent
    )
    click.echo(tabulate([('name', 'URL', 'ignore updates'), *installed_auras]))


@main.command(hidden=True)
@click.argument('filename', type=click.Path(dir_okay=False))
def generate_catalogue(filename: str):
    import asyncio
    from .resolvers import MasterCatalogue

    catalogue = asyncio.run(MasterCatalogue.collate())
    expanded = Path(filename)
    expanded.write_text(catalogue.json(indent=2), encoding='utf-8')

    compact = expanded.with_suffix(f'.compact{expanded.suffix}')
    compact.write_text(catalogue.json(separators=(',', ':')), encoding='utf-8')



import pytest
from yarl import URL

from instawow.wa_updater import ApiMetadata, AuraEntry, WaCompanionBuilder


@pytest.fixture
def builder(manager):
    yield WaCompanionBuilder(manager, 'test')


@pytest.mark.skip
@pytest.mark.asyncio
@pytest.mark.parametrize('ids', [['RaidCDs', 'bfaraid2'], ['bfaraid2', 'RaidCDs']])
async def test_id_order_is_retained_in_aura_metadata(builder, ids):
    results = await builder.get_wago_aura_metadata(ids)
    assert ids == [r.slug for r in results]


@pytest.mark.skip
@pytest.mark.asyncio
async def test_id_length_is_retained_in_aura_metadata(builder):
    ids = ['bfaraid2', 'foobar', 'RaidCDs']  # Invalid ID flanked by two valid IDs
    results = await builder.get_wago_aura_metadata(ids)
    assert [ApiMetadata, type(None), ApiMetadata] == [type(r) for r in results]


def test_can_parse_empty_displays_table(builder):
    assert (
        builder.extract_auras(
            '''\
WeakAurasSaved = {
    ["displays"] = {
    },
}
'''
        )
        == {}
    )


def test_urlless_display_is_discarded(builder):
    assert (
        builder.extract_auras(
            '''\
WeakAurasSaved = {
    ["displays"] = {
        ["Foo"] = {
            ["bar"] = "baz",
        },
    },
}
'''
        )
        == {}
    )


def test_can_parse_minimal_wago_display(builder):
    aura = AuraEntry(
        id='foo',
        uid='foo',
        parent=None,
        url=URL('https://wago.io/foo/1'),
        version=1,
        semver=None,
        ignore_wago_update=False,
    )
    assert (
        builder.extract_auras(
            '''\
WeakAurasSaved = {
    ["displays"] = {
        ["Foo"] = {
            ["bar"] = "baz",
            ["url"] = "https://wago.io/foo/1",
            ["version"] = 1,
            ["id"] = "foo",
            ["uid"] = "foo",
        },
    },
}
'''
        )
        == {'foo': [aura]}
    )


@pytest.mark.xfail
def test_url_host_not_wago_display_is_discarded(builder):
    assert (
        builder.extract_auras(
            '''\
WeakAurasSaved = {
    ["displays"] = {
        ["Foo"] = {
            ["bar"] = "baz",
            ["url"] = "https://wafo.io/foo/1",
            ["version"] = 1,
            ["id"] = "foo",
            ["uid"] = "foo",
        },
    },
}
'''
        )
        == {}
    )


def test_can_build_addon_without_updates(builder):
    builder.make_addon([])


def test_build_is_reproducible(builder):
    builder.make_addon([])
    checksum = builder.checksum()
    builder.make_addon([])
    assert checksum == builder.checksum()



from __future__ import annotations

import asyncio
from contextlib import asynccontextmanager
import contextvars as cv
from functools import partial
from itertools import filterfalse, starmap
import json
from pathlib import Path, PurePath
import posixpath
from shutil import copy, move
from tempfile import NamedTemporaryFile, mkdtemp
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncContextManager as ACM,
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    NoReturn,
    Optional as O,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
)

from loguru import logger

from . import DB_REVISION, exceptions as E
from .models import Pkg, PkgFolder, is_pkg
from .resolvers import (
    CurseResolver,
    Defn,
    InstawowResolver,
    MasterCatalogue,
    Resolver,
    TukuiResolver,
    WowiResolver,
)
from .utils import (
    bucketise,
    dict_chain,
    gather,
    is_not_stale,
    make_progress_bar,
    run_in_thread as t,
    shasum,
)

if TYPE_CHECKING:
    import aiohttp
    from prompt_toolkit.shortcuts import ProgressBar
    from sqlalchemy.orm import scoped_session
    from .config import Config

    _T = TypeVar('_T')
    _ArchiveR = Tuple[List[str], Callable[[Path], Awaitable[None]]]


USER_AGENT = 'instawow (https://github.com/layday/instawow)'

_web_client: cv.ContextVar[aiohttp.ClientSession] = cv.ContextVar('_web_client')

AsyncNamedTemporaryFile = t(NamedTemporaryFile)
amkdtemp = t(mkdtemp)
acopy = t(copy)
amove = t(move)


@asynccontextmanager
async def open_temp_writer() -> AsyncIterator[Tuple[Path, Callable[..., Any]]]:
    fh = await AsyncNamedTemporaryFile(delete=False)
    path = Path(fh.name)
    try:
        yield (path, t(fh.write))
    except BaseException:
        await t(fh.close)()
        await t(path.unlink)()
        raise
    else:
        await t(fh.close)()


async def trash(paths: Sequence[Path], parent: PurePath, *, missing_ok: bool = False) -> None:
    dst = await amkdtemp(dir=parent, prefix='deleted-' + paths[0].name + '-')
    for path in map(str, paths):  # https://bugs.python.org/issue32689
        try:
            await amove(path, dst)
        except (FileNotFoundError if missing_ok else ()):
            logger.opt(exception=True).info('source missing')


# macOS 'resource forks' are sometimes included in download zips - these
# violate our one add-on per folder contract-thing and will be omitted
_zip_excludes = {'__MACOSX'}


def find_base_dirs(names: Sequence[str]) -> Set[str]:
    return {
        n for n in (posixpath.dirname(n) for n in names) if n and posixpath.sep not in n
    } - _zip_excludes


def should_extract(base_dirs: Set[str]) -> Callable[[str], bool]:
    def is_member(name: str) -> bool:
        head, sep, _ = name.partition(posixpath.sep)
        return cast(bool, sep) and head in base_dirs

    return is_member


@asynccontextmanager
async def acquire_archive(path: PurePath) -> AsyncIterator[_ArchiveR]:
    from zipfile import ZipFile

    def extract(parent: Path) -> None:
        conflicts = base_dirs & {f.name for f in parent.iterdir()}
        if conflicts:
            raise E.PkgConflictsWithForeign(conflicts)
        else:
            members = filter(should_extract(base_dirs), names)
            archive.extractall(parent, members=members)

    archive = await t(ZipFile)(path)
    try:
        names = archive.namelist()
        base_dirs = find_base_dirs(names)
        yield (sorted(base_dirs), t(extract))
    finally:
        await t(archive.close)()


async def download_archive(
    manager: Manager, pkg: Pkg, *, chunk_size: int = 4096
) -> ACM[_ArchiveR]:
    url = pkg.download_url
    dst = manager.config.cache_dir / shasum(pkg.source, pkg.id, pkg.file_id)

    if await t(dst.exists)():
        pass
    elif url.startswith('file://'):
        from urllib.parse import unquote

        await acopy(unquote(url[7:]), dst)
    else:
        kwargs = {'raise_for_status': True, 'trace_request_ctx': {'show_progress': True}}
        async with manager.web_client.get(url, **kwargs) as response, open_temp_writer() as (
            temp_path,
            write,
        ):
            async for chunk in response.content.iter_chunked(chunk_size):
                await write(chunk)

        await amove(temp_path, dst)
    return acquire_archive(dst)


async def cache_json_response(manager: Manager, url: str, *args: Any, label: O[str] = None) -> Any:
    dst = manager.config.cache_dir / shasum(url)

    if await t(is_not_stale)(dst, *args):
        text = await t(dst.read_text)(encoding='utf-8')
    else:
        kwargs: Dict[str, Any] = {'raise_for_status': True}
        if label:
            kwargs.update({'trace_request_ctx': {'show_progress': True, 'label': label}})
        async with manager.web_client.get(url, **kwargs) as response:
            text = await response.text()

        await t(dst.write_text)(text, encoding='utf-8')
    return json.loads(text)


def prepare_db_session(config: Config) -> scoped_session:
    from sqlalchemy import create_engine
    from sqlalchemy.orm import sessionmaker, scoped_session
    from .models import ModelBase, should_migrate

    db_path = config.config_dir / 'db.sqlite'
    db_url = f'sqlite:///{db_path}'
    # We can't perform a migration on a new database without "metadata"
    # and we need to check whether it exists before calling ``create_engine``,
    # which will implicitly create the database file
    db_exists = db_path.exists()

    engine = create_engine(db_url)
    if should_migrate(engine, DB_REVISION):
        from alembic.command import stamp, upgrade
        from alembic.config import Config as AConfig
        from .utils import copy_resources

        with copy_resources(
            f'{__package__}.migrations', f'{__package__}.migrations.versions'
        ) as tmp_dir:
            aconfig = AConfig()
            aconfig.set_main_option('script_location', str(tmp_dir / __package__ / 'migrations'))
            aconfig.set_main_option('sqlalchemy.url', db_url)

            if db_exists:
                logger.info(f'migrating database to {DB_REVISION}')
                upgrade(aconfig, DB_REVISION)
            else:
                ModelBase.metadata.create_all(engine)
                logger.info(f'stamping database with {DB_REVISION}')
                stamp(aconfig, DB_REVISION)

    return scoped_session(sessionmaker(bind=engine))


def init_web_client(**kwargs: Any) -> aiohttp.ClientSession:
    from aiohttp import ClientSession, ClientTimeout, TCPConnector

    kwargs = {
        'connector': TCPConnector(force_close=True, limit_per_host=10),
        'headers': {'User-Agent': USER_AGENT},
        'trust_env': True,  # Respect http_proxy env var
        'timeout': cast(Any, ClientTimeout)(connect=15),
        **kwargs,
    }
    return ClientSession(**kwargs)


@object.__new__
class _DummyResolver:
    async def synchronise(self) -> _DummyResolver:
        return self

    async def resolve(self, defns: List[Defn], **kwargs: Any) -> Dict[Defn, E.PkgSourceInvalid]:
        return dict.fromkeys(defns, E.PkgSourceInvalid())


class _ResolverDict(dict):
    def __missing__(self, key: str) -> Resolver:
        return _DummyResolver


async def _error_out(error: Union[E.ManagerError, E.InternalError]) -> NoReturn:
    raise error


class Manager:
    config: Config
    db_session: Any  # scoped_session
    resolvers: Mapping[str, Resolver]
    catalogue: MasterCatalogue

    def __init__(self, config: Config, db_session: scoped_session) -> None:
        self.config = config
        self.db_session = db_session

        resolvers = (CurseResolver, WowiResolver, TukuiResolver, InstawowResolver)
        self.resolvers = _ResolverDict({r.source: r(self) for r in resolvers})
        self.catalogue = None  # type: ignore

    @property
    def web_client(self) -> aiohttp.ClientSession:
        return _web_client.get()

    @web_client.setter
    def web_client(self, value: aiohttp.ClientSession) -> None:
        _web_client.set(value)

    def get(self, defn: Defn) -> O[Pkg]:
        "Get a package from (source, id) or (source, slug)."
        return (
            self.db_session.query(Pkg)
            .filter(Pkg.source == defn.source, (Pkg.id == defn.name) | (Pkg.slug == defn.name))
            .first()
        )

    def get_from_substr(self, defn: Defn) -> O[Pkg]:
        "Get a package from a partial slug."
        return self.get(defn) or (
            self.db_session.query(Pkg)
            .filter(Pkg.slug.contains(defn.name))
            .order_by(Pkg.name)
            .first()
        )

    def decompose_url(self, url: str) -> O[Tuple[str, str]]:
        "Parse a URL into a source–name tuple."
        for resolver in self.resolvers.values():
            name = resolver.decompose_url(url)
            if name:
                return resolver.source, name

    async def synchronise(self) -> None:
        "Fetch the master catalogue from the interwebs."
        if self.catalogue is None:
            label = 'Synchronising master catalogue'
            url = (
                'https://raw.githubusercontent.com/layday/instascrape/data/'
                'master-catalogue-v1.compact.json'
            )  # v1
            catalogue = await cache_json_response(self, url, 4, 'hours', label=label)
            self.catalogue = MasterCatalogue.parse_obj(catalogue)

    async def _consume_seq(
        self, coros_by_defn: Dict[Defn, Callable[..., Awaitable[Any]]]
    ) -> Dict[Defn, E.ManagerResult]:
        return {d: await E.ManagerResult.acapture(c()) for d, c in coros_by_defn.items()}

    async def _resolve_deps(self, results: Iterable[Any]) -> Dict[Defn, Any]:
        """Resolve package dependencies.

        The resolver will not follow dependencies
        more than one level deep.  This is to avoid unnecessary
        complexity for something that I would never expect to
        encounter in the wild.
        """
        pkgs = list(filter(is_pkg, results))
        dep_defns = list(
            filterfalse(
                {(p.source, p.id) for p in pkgs}.__contains__,
                # Using a dict to maintain dep appearance order
                {(p.source, d.id): ... for p in pkgs for d in p.deps},
            )
        )
        if not dep_defns:
            return {}

        deps = await self.resolve(list(starmap(Defn, dep_defns)))
        pretty_deps = {d.with_name(r.slug) if is_pkg(r) else d: r for d, r in deps.items()}
        return pretty_deps

    async def resolve(self, defns: Sequence[Defn], with_deps: bool = False) -> Dict[Defn, Any]:
        "Resolve definitions into packages."
        if not defns:
            return {}

        await self.synchronise()
        defns_by_source = bucketise(defns, key=lambda v: v.source)

        results = await gather(self.resolvers[s].resolve(b) for s, b in defns_by_source.items())
        results_by_defn = dict_chain(defns, None, *(r.items() for r in results))
        if with_deps:
            results_by_defn.update(await self._resolve_deps(results_by_defn.values()))
        return results_by_defn

    async def search(self, search_terms: str, limit: int) -> Dict[Defn, Pkg]:
        "Search the combined names catalogue for packages."
        import heapq
        import string
        from jellyfish import jaro_winkler

        trans_table = str.maketrans(dict.fromkeys(string.punctuation, ' '))

        def normalise(value: str):
            return value.casefold().translate(trans_table).strip()

        await self.synchronise()

        s = normalise(search_terms)
        tokens_to_defns = bucketise(
            (
                (normalise(i.name), (i.source, i.id))
                for i in self.catalogue.__root__
                if self.config.game_flavour in i.compatibility
            ),
            key=lambda v: v[0],
        )

        # TODO: weigh matches under threshold against download count
        matches = heapq.nlargest(
            limit, ((jaro_winkler(s, n), n) for n in tokens_to_defns.keys()), key=lambda v: v[0]
        )
        defns = [Defn(*d) for _, m in matches for _, d in tokens_to_defns[m]]
        results = await self.resolve(defns)
        pkgs_by_defn = {d.with_name(r.slug): r for d, r in results.items() if is_pkg(r)}
        return pkgs_by_defn

    async def install_one(
        self, pkg: Pkg, archive: ACM[_ArchiveR], replace: bool
    ) -> E.PkgInstalled:
        async with archive as (folders, extract):
            conflicts = (
                self.db_session.query(Pkg)
                .join(Pkg.folders)
                .filter(PkgFolder.name.in_(folders))
                .all()
            )
            if conflicts:
                raise E.PkgConflictsWithInstalled(conflicts)

            if replace:
                await trash([self.config.addon_dir / f for f in folders], self.config.temp_dir)
            await extract(self.config.addon_dir)

        pkg.folders = [PkgFolder(name=f) for f in folders]
        self.db_session.add(pkg)
        self.db_session.commit()
        return E.PkgInstalled(pkg)

    async def install(self, defns: Sequence[Defn], replace: bool) -> Dict[Defn, E.ManagerResult]:
        prelim_results = await self.resolve([d for d in defns if not self.get(d)], with_deps=True)
        # Weed out installed deps - this isn't super efficient but avoids
        # having to deal with local state in the resolver
        results = {d: r for d, r in prelim_results.items() if not self.get(d)}
        installables = {(d, r): download_archive(self, r) for d, r in results.items() if is_pkg(r)}
        archives = await gather(installables.values())

        coros = dict_chain(
            defns,
            partial(_error_out, E.PkgAlreadyInstalled()),
            ((d, partial(_error_out, r)) for d, r in results.items()),
            (
                (d, partial(self.install_one, p, a, replace))
                for (d, p), a in zip(installables, archives)
            ),
        )
        return await self._consume_seq(coros)

    async def update_one(self, old_pkg: Pkg, pkg: Pkg, archive: ACM[_ArchiveR]) -> E.PkgUpdated:
        async with archive as (folders, extract):
            conflicts = (
                self.db_session.query(Pkg)
                .join(Pkg.folders)
                .filter(PkgFolder.pkg_source != pkg.source, PkgFolder.pkg_id != pkg.id)
                .filter(PkgFolder.name.in_(folders))
                .all()
            )
            if conflicts:
                raise E.PkgConflictsWithInstalled(conflicts)

            await self.remove_one(old_pkg)
            await extract(self.config.addon_dir)

        pkg.folders = [PkgFolder(name=f) for f in folders]
        self.db_session.add(pkg)
        self.db_session.commit()
        return E.PkgUpdated(old_pkg, pkg)

    async def update(self, defns: Sequence[Defn]) -> Dict[Defn, E.ManagerResult]:
        # Rebuild ``Defn`` with ID and strategy from package for defns
        # of installed packages.  Using the ID has the benefit of resolving
        # installed-but-renamed packages
        maybe_pkgs = ((d, self.get(d)) for d in defns)
        checked_defns = {Defn.from_pkg(p) if p else c: p for c, p in maybe_pkgs}

        # Results can contain errors
        # installables are those results which are packages
        # and updatables are packages with updates
        results = await self.resolve([d for d, p in checked_defns.items() if p])
        installables = {d: r for d, r in results.items() if is_pkg(r)}
        updatables = {
            (d, checked_defns[d], p): download_archive(self, p)
            for d, p in installables.items()
            if p.file_id != cast(Pkg, checked_defns[d]).file_id
        }
        archives = await gather(updatables.values())

        coros = dict_chain(
            checked_defns,
            partial(_error_out, E.PkgNotInstalled()),
            ((d, partial(_error_out, r)) for d, r in results.items()),
            ((d, partial(_error_out, E.PkgUpToDate())) for d in installables),
            ((d, partial(self.update_one, *p, a)) for (d, *p), a in zip(updatables, archives)),
        )
        return await self._consume_seq(coros)

    async def remove_one(self, pkg: Pkg) -> E.PkgRemoved:
        await trash(
            [self.config.addon_dir / f.name for f in pkg.folders],
            parent=self.config.temp_dir,
            missing_ok=True,
        )
        self.db_session.delete(pkg)
        self.db_session.commit()
        return E.PkgRemoved(pkg)

    async def remove(self, defns: Sequence[Defn]) -> Dict[Defn, E.ManagerResult]:
        pkgs_by_defn = ((d, self.get(d)) for d in defns)
        coros = dict_chain(
            defns,
            partial(_error_out, E.PkgNotInstalled()),
            ((d, partial(self.remove_one, p)) for d, p in pkgs_by_defn if p),
        )
        return await self._consume_seq(coros)


_tick_interval = 0.1
_tickers: cv.ContextVar[Set[asyncio.Task[None]]] = cv.ContextVar('_tickers', default=set())


@asynccontextmanager
async def cancel_tickers() -> AsyncIterator[None]:
    try:
        yield
    finally:
        for ticker in _tickers.get():
            ticker.cancel()


def init_cli_web_client(*, Bar: ProgressBar) -> aiohttp.ClientSession:
    from cgi import parse_header
    from aiohttp import TraceConfig, hdrs

    def extract_filename(response: aiohttp.ClientResponse) -> str:
        _, cd_params = parse_header(response.headers.get(hdrs.CONTENT_DISPOSITION, ''))
        filename = cd_params.get('filename') or response.url.name
        return filename

    async def do_on_request_end(session: Any, request_ctx: Any, params: Any) -> None:
        ctx = request_ctx.trace_request_ctx
        if not (ctx and ctx.get('show_progress')):
            return

        async def ticker() -> None:
            label = ctx.get('label') or f'Downloading {extract_filename(params.response)}'
            total = params.response.content_length
            if (
                total is None
                # Size before decoding is not exposed in streaming API and
                # `Content-Length` has the size of the payload after gzipping
                or params.response.headers.get(hdrs.CONTENT_ENCODING) == 'gzip'
            ):
                # Length of zero will have a hash sign cycle through the bar
                # (see indeterminate progress bars)
                total = 0

            bar = None
            try:
                bar = Bar(label=label, total=total)
                content = params.response.content

                while not content.is_eof():
                    # This is ``bar.current`` in prompt_toolkit v2
                    # and ``.items_completed`` in v3
                    bar.current = bar.items_completed = content.total_bytes
                    Bar.invalidate()
                    await asyncio.sleep(_tick_interval)
            finally:
                if bar is not None:
                    Bar.counters.remove(bar)

        tickers = _tickers.get()
        tickers.add(asyncio.create_task(ticker()))

    trace_config = TraceConfig()
    trace_config.on_request_end.append(do_on_request_end)
    trace_config.freeze()
    return init_web_client(trace_configs=[trace_config])


class CliManager(Manager):
    def run(self, awaitable: Awaitable[_T]) -> _T:
        async def run():
            async with init_cli_web_client(Bar=Bar) as self.web_client, cancel_tickers():
                return await awaitable

        with make_progress_bar() as Bar:
            loop = asyncio.new_event_loop()
            try:
                return loop.run_until_complete(run())
            finally:
                loop.close()



import os
import subprocess as sp

import nox

nox.options.envdir = '.py-nox'


@nox.session(python=['3.7', '3.8'])
def test(session):
    session.install('.[test]')
    session.run('coverage', 'run', '-m', 'pytest', '-o', 'xfail_strict=true', 'tests')
    session.run('coverage', 'report', '-m')


@nox.session(python=['3.7', '3.8'])
def type_check(session):
    session.install('.')
    session.run('npx', '--cache', '.npm', 'pyright', '--lib')


@nox.session(python='3.7')
def reformat(session):
    session.install('isort[pyproject]>=4.3.5', 'black')
    session.run('isort', '--recursive', 'instawow', 'tests', 'noxfile.py', 'setup.py')
    session.run(
        'black', '--config', 'pyproject.toml', 'instawow', 'tests', 'noxfile.py', 'setup.py'
    )


@nox.session(python=False)
def update_stubs(session):
    sp.run(
        """\
        types_dir=.py-types
        rm -rf $types_dir && mkdir $types_dir && cd $types_dir && {
          git clone --depth 1 \
            https://github.com/python/typeshed
          git clone --depth 1 \
            https://github.com/dropbox/sqlalchemy-stubs stubs/_sqlalchemy-stubs
          cp -r stubs/_sqlalchemy-stubs/sqlalchemy-stubs stubs/sqlalchemy
        }
    """,
        shell=True,
        executable='bash',
    )


@nox.session(python=False)
def clobber_build_artefacts(session):
    session.run('rm', '-rf', 'build', 'dist', 'instawow.egg-info')


@nox.session(python='3.7')
def build(session):
    clobber_build_artefacts(session)
    session.install('pep517')
    session.run('python', '-m', 'pep517.build', '.')


@nox.session(python='3.7')
def publish(session):
    session.install('twine')
    session.run('twine', 'upload', 'dist/*')


@nox.session
def nixify(session):
    session.cd(os.environ.get('INSTAWOW_NIXIFY_DIR', '.'))
    session.install('pypi2nix')
    session.run('pypi2nix', '-e', 'instawow')



import pytest

from instawow.resolvers import Strategies

strategy = Strategies.default


@pytest.mark.asyncio
async def test_search(manager):
    results = await manager.search('molinari', limit=5)
    assert ('curse', 'molinari', strategy, None) in results and (
        'wowi',
        '13188-molinari',
        strategy,
        None,
    ) in results


@pytest.mark.asyncio
async def test_search_with_extra_spice(manager):
    results = await manager.search('AtlasLootClassic', limit=5)
    if manager.config.is_classic:
        assert ('curse', 'atlaslootclassic', strategy, None) in results
    else:
        assert ('curse', 'atlaslootclassic', strategy, None) not in results
